//------------------------------------------------------------------
//  Batch mode application for the OPN2002
//------------------------------------------------------------------
//  This demo allows the OPN2002 to read barcodes, store them and
//  transmit them by Neto using USB or the CRD-2000 multicradle.
//
//  This application also support the reading of the
//  univeral menu-book, so all sorts of barcode options can easily be tested.
//
//  Since all changed configurations are stored in flash, any changed
//  configurations are restored after a restart or when the battery was empty.
//
//  See the application manual on more details on the support
//  features and on how the application works.
//------------------------------------------------------------------
//  Version history:
//	RFx35410		:	First release
//                      Minimal required OS version: RBGV0207
//						Release date: December 7th, 2009
//						Author: Ronny de Winter
//						Company: Opticon Sensors Europe BV
//
//	RFG35411		:	Added configurable database formatting
//						Minimal required OS version: RBGV0207
//						Release date: December 14, 2009
//						Author: Ronny de Winter
//						Company: Opticon Sensors Europe BV
//
//	RFG35412		:	Fixed very slow storing of barcodes
//						Minimal required OS version: RBGV0207
//						Release date: February 3, 2010
//						Author: Ronny de Winter
//						Company: Opticon Sensors Europe BV
//
//	RFG35413		:	Preliminary addition of USB-MSD
//						Minimal required OS version: RBxV0220b
//						Release date: June 10, 2011
//						Author: Michiel van den Brink
//						Company: Opticon Sensors Europe BV
//
//	RFG35414		:	Added USB-MSD and updated manual
//						Minimal required OS version: RBxV0220c
//						Release date: June 17, 2011
//						Author: Ronny de Winter
//						Company: Opticon Sensors Europe BV
//
//	RFG35415		:	Fixed USB-MSD issue
//						Minimal required OS version: RBxV0220d
//						Release date: June 29, 2011
//						Author: Ronny de Winter
//						Company: Opticon Sensors Europe BV
//
//	RFG35416		:	Disable flashing green LED indicator and re-enable charge indicator
//						Minimal required OS version: RBxV0221
//						Release date: May 21, 2012
//						Author: Chris Parker
//						Company: Opticon Sensors Europe BV
//
//	RFG35415		:	Add OseComm
//						Release date: April 29, 2013
//						Minimal required OS version: RBxV0221b
//						Author: Ben Hartman
//						Company: Opticon Sensors Europe BV
//
//------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "lib.h"
#include "RFG3541.h"
#include "database.h"
//
#include "menu_options.inc"		// Contains table with all application specific menu options
//

// Defines and variables used for barcode reading
#define	MAX_BARCODE_LENGTH		99

// Database name
#define DBASE_NAME	   	"DATA.TXT"

// Error values
#define ILLEGAL_QUANTITY	0xFFFF
#define DATABASE_ERROR		0xFFFE
#define FILESYSTEM_ERROR	0xFFFD
#define ERROR_VALUE			0xFFF0



// Database record
#define SZ_BARCODE		(MAX_BARCODE_LENGTH+1)
#define SZ_QUANTITY		9
#define SZ_SERIAL		6
#define SZ_TIME			(2+1+2+1+2)
#define SZ_DATE			(4+1+2+1+2)
#define SZ_RECORD		(SZ_BARCODE+1+SZ_QUANTITY+1+SZ_SERIAL+1+SZ_TIME+1+SZ_DATE+1+1)

// barcode record structure
typedef struct
{
	char barcode[ SZ_BARCODE + 1 ];
	char quantity[ SZ_QUANTITY + 1 ];
	char serial[SZ_SERIAL + 1 ];
	char time[ SZ_TIME + 1 ];
	char date[ SZ_DATE + 1 ];
}db_record;


static char bcr_buf[MAX_BARCODE_LENGTH];
static struct barcode code = {0};

#define IDLE_POWER_DOWN_TIME		    3*50		// Powerdown time when idle: 3 seconds

//
// States of the main state machine of this application
//
#define STATE_IDLE				 0		// Idle state
#define STATE_RS232_CONNECTED	 1		// Placed in Cradle
#define STATE_USB_CONNECTED		 2		// Connected to USB
#define STATE_WAIT_FOR_RELEASE	 3		// Wait until both keys are not pressed and then jump back idle state

//
// Public functions
//
void Application_Default(void);
void Version_Output(void);
int FieldSize(int field_type);

#define	VERSION_CHECK		"RFG3541.3"			// Verification string used by 'LoadApplicationSettings' to check if any possible 
												// application settings found in flash memory are compatible with this application.
												// If this version check fails all application settings will be set back to default
												// Note: Only change this version if the application settings of the previous version are no longer compatible

application_vars app;							// Structure containing all application settings that will be stored in flash memory

static int last_key_pressed = -1;				// Used to determine whether a barcode was read using the clear key or trigger key

//------------------------------------------------------------------
//  Application_Default
//  ============================
//  Reset all application specific settings to default
//------------------------------------------------------------------
void Application_Default(void)
{
	systemsetting("U2");											// OSE default
	systemsetting("YC");											// Enable menu-labels
	systemsetting("7G");											// Enable charge indicator

	app.default_interface = COM0;									// Interface: USB-VCP / Cradle (RS232)
	app.read_mode = SINGLE;											// Single read
	app.read_time = 0;												// Trigger mode
	app.goodread_duration = 10;										// Good read led duration (200ms)
	app.buzzer_volume = VHIGH;										// Buzzer volume (high)
	app.buzzer_duration = TSTANDARD;								// Buzzer duration (200ms)
	app.buzzer_options = BUZZER_ENABLED | STARTUP_BUZZER;			// Buzzer enabled, startup buzzer enabled
	app.buzzer_options |= BUZZER_TONE_HIGH_LOW;						// Buzzer tone (High-low tone)
	app.special_options = 0;										// RFU (Reserved for future use)

	app.output_field_separator = ',';								// Output field separator
	app.output_date_time_format = TIME_FORMAT_HH_MM_SS;				// Output time format hh:mm:ss
	app.output_date_time_format |= DATE_FORMAT_SLASH_DD_MM_YYYY;	// Output data dd/mm/yyyy
	app.output_barcode_size = 0x40;									// Output barcode field size (40 characters, stored as BCD value)
	app.output_quantity_size = 6;									// Output quantity field size (0=empty field)
	app.output_quantity_options = QUANTITY_DELETE_ZERO_QNTY;		// Don't allow negative quantity, delete key enabled, delete record is quantity reaches zero

	app.output_sequence[0] = SEQUENCE_BARCODE;						// Output sequence: Barcode, Quantity, Serial number, Time, Date
	app.output_sequence[1] = SEQUENCE_QUANTITY;
	app.output_sequence[2] = SEQUENCE_SERIAL_NUMBER;
	app.output_sequence[3] = SEQUENCE_TIME;
	app.output_sequence[4] = SEQUENCE_DATE;
}

//------------------------------------------------------------------
//  Version_Output
//  ============================
//  Output all versions, including the application version
//------------------------------------------------------------------
void Version_Output(void)
{
	systemsetting("Z1");

	printf("%s\r", VERSION);
}

//------------------------------------------------------------------
//  OnOseComInfo
//  ============================
//  Delete file when succesfull download is completed
//------------------------------------------------------------------

void OnOseComInfo(int status, int errorsuccess, int progress, const char *info)
{
    switch (status)
    {
    case STAT_GET_TIME_DATE:
        break;
    case STAT_SET_TIME_DATE:
        break;
    case STAT_GET_OS_VERSION:
        break;
    case STAT_GET_APPL_VERSION:
        break;
    case STAT_XMIT_FILE_TO_PC:
        // if successful then delete the file
        if (errorsuccess == SUCC_COMPLETE)
            remove(info);
        break;
    case STAT_RECV_FILE_FROM_PC:
        break;
    case STAT_LIST_FILES:
        break;
    }
}

//------------------------------------------------------------------
//  TransmitData
//  ============================
//  Tries to transmit the database file using the Neto or OseComm protocol
//------------------------------------------------------------------
int TransmitData(void)
{
#ifdef OSECOMM
	//
	if (OseComm( -1L, CRADLE_ABORT, "OseComm", OnOseComInfo) == COMM_OK)
	{
		sound( TSTANDARD, VHIGH, SMEDIUM, SHIGH, SLOW, SMEDIUM, SHIGH, 0 );		// Succes beep
		return OK;
	}
	else
	{
		return -1;
	}
#else	// #ifdef NETO
	char serial_number[6+1];
	int result;

	if( fsize((char*)DBASE_NAME) == -1L )
	{
		return OK;
	}

	sprintf(serial_number, "%-6.6lu", GetSerialNumber());

	result = neto_transmit( (char(*)[MAX_FNAME])DBASE_NAME, 1, serial_number, TRIGGER_KEY, 3);

	if(result >= 0)
	{
		remove(DBASE_NAME);
		sound( TSTANDARD, VHIGH, SMEDIUM, SHIGH, SLOW, SMEDIUM, SHIGH, 0 );		// Succes beep
		return OK;
	}

	return -1;
#endif
}

//------------------------------------------------------------------
//  StoreBarcodeData
//  ============================
//  Saves the barcode data into the database file
//------------------------------------------------------------------
int StoreBarcodeData( db_record *db_rec, long lRecordNo, int delete_record )
{
	static SDBFile dbFile; // static initializes all items to 0
	static char record[ SZ_RECORD + 1 ];

	char *ptr;
	int i, field_size, record_size;
	int barcode_offset, barcode_size, quantity_offset;

	// Check if there is enough space available for storing the barcode data
	if( coreleft() < 1024 )
	{
		return FILESYSTEM_ERROR;
	}

	ptr = record;
	record_size = barcode_size = 0;
	barcode_offset = quantity_offset = -1;

	for(i=0; i<5; ++i)
	{
		field_size = FieldSize(app.output_sequence[i]);

		switch(app.output_sequence[i])
		{
			case SEQUENCE_BARCODE:	
				barcode_size = field_size;
				barcode_offset = record_size;
				sprintf(ptr, "%-*.*s", field_size, field_size, db_rec->barcode);
				break;

			case SEQUENCE_QUANTITY:
				quantity_offset = record_size;
				sprintf(ptr, "%-*.*s", field_size, field_size, db_rec->quantity);
				break;

			case SEQUENCE_SERIAL_NUMBER:
				sprintf(ptr, "%-*.*s", field_size, field_size, db_rec->serial);
				break;

			case SEQUENCE_TIME:
				sprintf(ptr, "%-*.*s", field_size, field_size, db_rec->time);
				break;

			case SEQUENCE_DATE:
				sprintf(ptr, "%-*.*s", field_size, field_size, db_rec->date);
				break;
		}

		record_size += field_size;

		ptr += field_size;						// Skip the field
		
		if(i==5-1 || app.output_sequence[i+1] == SEQUENCE_NONE || app.output_sequence[i+1] > SEQUENCE_MAX)
		{
			strcpy(ptr, "\r\n");
			record_size += 2;					// Add 2 for end-of-record separator
			break;
		}

		*ptr++ = app.output_field_separator;	// Add field separator

		++record_size;							// +1 for field separator
	}

	if( !OpenDatabase( (char*)DBASE_NAME, record_size, &dbFile ))
	{
		if( !CreateDatabase( (char*)DBASE_NAME, record_size, &dbFile ))
		{
			return FILESYSTEM_ERROR;
		}
	}

	if(delete_record)
	{
		if(lRecordNo==-1L || !DeleteRecord( &dbFile, lRecordNo) )
		{
			CloseDatabase( &dbFile );
			return DATABASE_ERROR;
		}
	}
	else
	{
		if( lRecordNo != -1L )
			GotoRecord( &dbFile, lRecordNo );

		if( !WriteRecord( &dbFile, record, ((lRecordNo==-1L) ? WRITE_APPEND : WRITE_OVER)))
		{
			CloseDatabase( &dbFile );
			return FILESYSTEM_ERROR;
		}
	}
	
	CloseDatabase( &dbFile );
	return OK;
}

//------------------------------------------------------------------
//  StringQuantityToLong
//  ============================
//  Convert the stored quantity string to a long value
//------------------------------------------------------------------
long StringQuantityToLong( char* quantity)
{
	static char tmp[SZ_QUANTITY+1];
	char* ptr;
	int pos;

	memset( tmp, '\0', sizeof( tmp ));
	ptr = tmp;
	for( pos = 0; pos < (SZ_QUANTITY); pos++ )
	{
		if( quantity[pos] == '\0' )
			break;

		if( quantity[pos] == ' ') // cut off all spaces
		 	continue;

		 *ptr++ = quantity[pos];
	}
	if( atol( tmp ) == 0L && tmp[0] == '-' )
		return 0;

	return( atol( tmp ));
}

//------------------------------------------------------------------
//  FieldSize
//  ============================
//  Returns the string length of the specified field type
//------------------------------------------------------------------
int FieldSize(int field_type)
{
	switch(field_type)
	{
		case SEQUENCE_BARCODE:	
			return (app.output_barcode_size>>4)*10 + (app.output_barcode_size & 0xF);

		case SEQUENCE_QUANTITY:
			return app.output_quantity_size;

		case SEQUENCE_SERIAL_NUMBER:
			return SZ_SERIAL;

		case SEQUENCE_TIME:
			if( (app.output_date_time_format & TIME_FORMAT_MASK) == TIME_FORMAT_HH_MM )
				return 5;

			return 8;		// TIME_FORMAT_HH_MM_SS:

		case SEQUENCE_DATE:

			switch(app.output_date_time_format & DATE_FORMAT_MASK)
			{
				case DATE_FORMAT_DASH_DD_MM_YYYY:	
				case DATE_FORMAT_DASH_MM_DD_YYYY:
				case DATE_FORMAT_DASH_YYYY_MM_DD:
				case DATE_FORMAT_SLASH_DD_MM_YYYY:	
				case DATE_FORMAT_SLASH_MM_DD_YYYY:
				case DATE_FORMAT_SLASH_YYYY_MM_DD:
					return 10;

				case DATE_FORMAT_DASH_DD_MM_YY:	
				case DATE_FORMAT_DASH_MM_DD_YY:		
				case DATE_FORMAT_DASH_YY_MM_DD:
				case DATE_FORMAT_SLASH_DD_MM_YY:	
				case DATE_FORMAT_SLASH_MM_DD_YY:
				case DATE_FORMAT_SLASH_YY_MM_DD:
					return 8;

				case DATE_FORMAT_DASH_DD_MM:
				case DATE_FORMAT_DASH_MM_DD:
				case DATE_FORMAT_SLASH_DD_MM:
				case DATE_FORMAT_SLASH_MM_DD:
					return 5;
			}
	}

	return 0;
}



//------------------------------------------------------------------
//  FillRecordStruct
//  ============================
//  Copies the fixed length record into a database record structure
//------------------------------------------------------------------
void FillRecordStruct( db_record *db_rec, char* record )
{
	char *ptr;
	int i, field_size;

	memset( db_rec, '\0', sizeof( db_record ));

	ptr = record;

	for(i=0; i<5; ++i)
	{
		field_size = FieldSize(app.output_sequence[i]);

		switch(app.output_sequence[i])
		{
			case SEQUENCE_BARCODE:	
				sprintf(db_rec->barcode, "%-*.*s", field_size, field_size, ptr);
				break;

			case SEQUENCE_QUANTITY:
				sprintf(db_rec->quantity, "%-*.*s", field_size, field_size, ptr);
				break;

			case SEQUENCE_SERIAL_NUMBER:
				sprintf(db_rec->serial, "%-*.*s", field_size, field_size, ptr);
				break;

			case SEQUENCE_TIME:
				sprintf(db_rec->time, "%-*.*s", field_size, field_size, ptr);
				break;

			case SEQUENCE_DATE:
				sprintf(db_rec->date, "%-*.*s", field_size, field_size, ptr);
				break;
		}

		ptr += field_size + 1;	// Skip the field + field separator

		if(i==5-1 || app.output_sequence[i+1] == SEQUENCE_NONE || app.output_sequence[i+1] > SEQUENCE_MAX)
			break;
	}
}

//-------------------------------------------------------------------
//  FindBarcodeInDatabase
//  ============================
//  Searches for the specified barcode in the database and returns it
//-------------------------------------------------------------------
long FindBarcodeInDatabase( char *barcode, char *quantity )
{
	static SDBFile dbFile;
	static char record[ SZ_RECORD + 1 ];
	static db_record db_rec;
	long lFound = -1L;
	int i, field_size, record_size, barcode_size;
	int barcode_offset, quantity_offset;

	barcode_size = record_size = 0;
	barcode_offset = quantity_offset = -1;

	for(i=0;i<5;++i)
	{
		field_size = FieldSize(app.output_sequence[i]);

		if(app.output_sequence[i] == SEQUENCE_BARCODE)
		{
			barcode_size = field_size;
			barcode_offset = record_size;
		}
		else if(app.output_sequence[i] == SEQUENCE_QUANTITY)
		{
			quantity_offset = record_size;
		}
		
		record_size += field_size;

		if(i==5-1 || app.output_sequence[i+1] == SEQUENCE_NONE || app.output_sequence[i+1] > SEQUENCE_MAX)
		{
			record_size += 2;		// Add 2 for end-of-record separator
			break;
		}

		++record_size;	// Add 1 for field separator
	}

	if(barcode_offset == -1)
		return -1L;

	if( !OpenDatabase( (char*)DBASE_NAME, record_size, &dbFile ))
		return -1L;

	if(quantity_offset == -1 || app.output_quantity_size == 0)
	{
		if(last_key_pressed != CLR_KEY)		// Little bit ugly speed hack to avoid doing a slow database search when it's not even needed
		{
			CloseDatabase( &dbFile );
			return -1L;						// Return -1, so record will be appended
		}

		if( (lFound = LineairSearch( &dbFile, record, barcode, barcode_size, barcode_offset )) != -1L )
		{
			// Barcode was found fill the record
			FillRecordStruct( &db_rec, record );
		}
	}
	else
	{
		if( (lFound = LineairSearch( &dbFile, record, barcode, barcode_size, barcode_offset )) != -1L )
		{
			// Barcode was found fill the quantity string
			FillRecordStruct( &db_rec, record );
			strncpy( quantity, db_rec.quantity, app.output_quantity_size );
		}
	}

	CloseDatabase( &dbFile );
	return lFound;
}


//------------------------------------------------------------------
//  StoreBarcode
//  ============================
//  Stores a barcode in the database
//------------------------------------------------------------------
int StoreBarcode(struct barcode *code, int amount)
{
	static char 		quantity[ SZ_QUANTITY + 1 ];
	static db_record	db_rec;
    struct date 		dates;
    struct time 		times;
    long   				lFoundRecord;
    long 				lTotal;
	int					do_delete = FALSE, result;
	int					barcode_size = FieldSize(SEQUENCE_BARCODE);

	memset( &db_rec, '\0', sizeof( db_record ));

	// fill the barcode into the record structure
	sprintf( db_rec.barcode, "%-*.*s", barcode_size, barcode_size, code->text );
	
	memset( quantity, '\0', sizeof( quantity ));	// clear the whole quantity item

	if( (lFoundRecord = FindBarcodeInDatabase( db_rec.barcode, quantity )) != -1L )
	{
		if(strlen(quantity) == 0 && amount < 0)		// If a record was found, but no quantity field
		{
			do_delete = TRUE;
			lTotal = 0;
		}
		else
		{
			lTotal = StringQuantityToLong(quantity);

			if( !(app.output_quantity_options & QUANTITY_ALLOW_NEGATIVE_QNTY) )
			{
				if(lTotal == 0 && amount < 0)
					return ILLEGAL_QUANTITY;
			}

			lTotal += amount;

			if(lTotal == 0 && (app.output_quantity_options & QUANTITY_DELETE_ZERO_QNTY) )
			{
				do_delete = TRUE;
			}
		}
	}
	else
	{
		if( !(app.output_quantity_options & QUANTITY_ALLOW_NEGATIVE_QNTY) )
		{
			if(amount < 0)
				return ILLEGAL_QUANTITY;
		}

		lTotal = amount;
	}

	// When ok fill the quantity of the db_record
	sprintf( db_rec.quantity, "%*ld", app.output_quantity_size, lTotal);

	sprintf( db_rec.serial, "%-6.6lu", GetSerialNumber() );

	gettime( &times );

	switch(app.output_date_time_format & TIME_FORMAT_MASK)
	{
		case TIME_FORMAT_HH_MM:
			sprintf( db_rec.time, "%02d:%02d", times.ti_hour, times.ti_min);
			break;
				
		case TIME_FORMAT_HH_MM_SS:
			sprintf( db_rec.time, "%02d:%02d:%02d", times.ti_hour, times.ti_min, times.ti_sec);
			break;
	}

	getdate( &dates );

	switch(app.output_date_time_format & DATE_FORMAT_MASK)
	{
		case DATE_FORMAT_SLASH_DD_MM_YYYY:
			sprintf( db_rec.date, "%02d/%02d/%04d", dates.da_day, dates.da_mon, dates.da_year);		break;
		case DATE_FORMAT_DASH_DD_MM_YYYY:
			sprintf( db_rec.date, "%02d-%02d-%04d", dates.da_day, dates.da_mon, dates.da_year);		break;
		case DATE_FORMAT_SLASH_MM_DD_YYYY:
			sprintf( db_rec.date, "%02d/%02d/%04d", dates.da_mon, dates.da_day, dates.da_year);		break;
		case DATE_FORMAT_DASH_MM_DD_YYYY:
			sprintf( db_rec.date, "%02d-%02d-%04d", dates.da_mon, dates.da_day, dates.da_year);		break;
		case DATE_FORMAT_SLASH_YYYY_MM_DD:
			sprintf( db_rec.date, "%04d/%02d/%02d", dates.da_year, dates.da_mon, dates.da_day);		break;
		case DATE_FORMAT_DASH_YYYY_MM_DD:
			sprintf( db_rec.date, "%04d-%02d-%02d", dates.da_year, dates.da_mon, dates.da_day);		break;
		case DATE_FORMAT_DASH_DD_MM_YY:	
			sprintf( db_rec.date, "%02d-%02d-%02d", dates.da_day, dates.da_mon, dates.da_year % 100);	break;
		case DATE_FORMAT_SLASH_DD_MM_YY:
			sprintf( db_rec.date, "%02d/%02d/%02d", dates.da_day, dates.da_mon, dates.da_year % 100);	break;
		case DATE_FORMAT_SLASH_MM_DD_YY:		
			sprintf( db_rec.date, "%02d/%02d/%02d", dates.da_mon, dates.da_day, dates.da_year % 100);	break;
		case DATE_FORMAT_DASH_MM_DD_YY:
			sprintf( db_rec.date, "%02d-%02d-%02d", dates.da_mon, dates.da_day, dates.da_year % 100);	break;	
		case DATE_FORMAT_SLASH_YY_MM_DD:
			sprintf( db_rec.date, "%02d/%02d/%02d", dates.da_year % 100, dates.da_mon, dates.da_day);	break;
		case DATE_FORMAT_DASH_YY_MM_DD:
			sprintf( db_rec.date, "%02d-%02d-%02d", dates.da_year % 100, dates.da_mon, dates.da_day);	break;
		case DATE_FORMAT_SLASH_MM_DD:
			sprintf( db_rec.date, "%02d/%02d", dates.da_mon, dates.da_day);	break;
		case DATE_FORMAT_DASH_MM_DD:
			sprintf( db_rec.date, "%02d-%02d", dates.da_mon, dates.da_day);	break;
		case DATE_FORMAT_SLASH_DD_MM:
			sprintf( db_rec.date, "%02d/%02d", dates.da_day, dates.da_mon);	break;
		case DATE_FORMAT_DASH_DD_MM:
			sprintf( db_rec.date, "%02d-%02d", dates.da_day, dates.da_mon);	break;
	}
	// Store the input data into our database
	result = StoreBarcodeData( &db_rec, lFoundRecord, do_delete);

	if(result != OK)
		return result;

	return (int)lTotal;
}

//------------------------------------------------------------------
//  SetReadMode
//  ============================
//  Converts the configured read time/mode to a scannerpower call
//------------------------------------------------------------------
void SetReadMode(int readmode, int readtime)
{
	switch(readmode)
	{
		case MULTIPLE:
			if(readtime == 0)
				scannerpower(MULTIPLE | TRIGGER, 10*50);		// Limit trigger mode to 10 seconds to avoid pressing the trigger for 20s, causing an unintended restart
			else if(readtime == READ_TIME_INDEFINITELY)
				scannerpower(MULTIPLE, -1);
			else
				scannerpower(MULTIPLE, readtime*5);
			break;

		case CONTINUOUS:
			if(readtime == 0)
				scannerpower(CONTINUOUS | TRIGGER, 10*50);		// Limit trigger mode to 10 seconds to avoid pressing the trigger for 20s, causing an unintended restart
			else if(readtime == READ_TIME_INDEFINITELY)
				scannerpower(CONTINUOUS, -1);
			else
				scannerpower(CONTINUOUS, readtime*5);
			break;

		case SINGLE:
		default:
			if(readtime == 0)
				scannerpower(SINGLE | TRIGGER, 10*50);			// Limit trigger mode to 10 seconds to avoid pressing the trigger for 20s, causing an unintended restart
			else if(readtime == READ_TIME_INDEFINITELY)
				scannerpower(SINGLE, -1);
			else
				scannerpower(SINGLE, readtime*5);
			break;
	}

	// If TRUE, then the small clear key acts like a trigger key, so read modes settings also apply to this key
	ClearKeyAsTrigger( (app.output_quantity_options & QUANTITY_DELETE_KEY_DISABLED) ? FALSE : TRUE );		
}

//------------------------------------------------------------------
//  OkBeep
//  ============================
//  Sounds a good read beep according to the current configuration
//------------------------------------------------------------------
void OkBeep(int base_frequency)
{
	unsigned char length, volume;

	if(!(app.buzzer_options & BUZZER_ENABLED))
		return;

	volume = app.buzzer_volume;
	length = app.buzzer_duration;

	switch( app.buzzer_options & BUZZER_TONE_MASK )
	{
		case BUZZER_TONE_SINGLE:
			sound( length * 2, volume, base_frequency, 0);
			break;

		case BUZZER_TONE_LOW_HIGH:
			sound( length, volume, base_frequency, base_frequency+1000, 0);
			break;

		case BUZZER_TONE_HIGH_LOW:
		default:
			sound( length, volume, base_frequency, base_frequency-500, 0);
			break;
	}
}

//------------------------------------------------------------------
//  ErrBeep
//  ============================
//  Sounds an error beep according to the current configuration
//------------------------------------------------------------------
void ErrBeep(void)
{
	unsigned char length, volume;

	volume = app.buzzer_volume;
	length = app.buzzer_duration;

	sound( length * 2, volume, SERROR, 0);
}

//------------------------------------------------------------------
//  GoodReadLed
//  ============================
//  Shows a good read LED according to the current configuration
//------------------------------------------------------------------
void GoodReadLed(int color)
{
	goodreadled(color, app.goodread_duration);
}

//------------------------------------------------------------------
//  ProcessBarcode
//  =====================
//  Processes a barcode that has been read. This function
//  makes is possible to read set-up labels from the universal
//  book as well. (Changed settings will be lost after a restart)
//------------------------------------------------------------------
int ProcessBarcode(struct barcode *code)
{
	int result = 0;

	if(code->id == MENU_CODE)		// Check for standard Code-39 menu labels (from the OSE Universal menu book)
	{
		scannerpower(ON, -1);		// Turn laser on until menu mode is exited

		while(result != EXITING_MENU_MODE && result != ERROR && !scanneroff())
		{
			if(result == 0 || readbarcode(code) == OK)
			{
				switch( (result=ExecuteMenuLabel(code)) )
				{
					case ENTERING_MENU_MODE:
						sound(TSTANDARD, VHIGH, SHIGH, SMEDIUM, SHIGH, 0);
						break;

					case EXITING_MENU_MODE:
						scannerpower(OFF, 0);
						sound(TSTANDARD,VHIGH, SHIGH, SMEDIUM, SHIGH, 0);
						delay(TSTANDARD * 4);		// Wait till buzzer sound if finished before saving (saving flash settings turns off sound timer temporarily)
						systemsetting("Z2");		// Save changed settings in flash memory

						if( comopen(-1) != app.default_interface )
						{
							comclose( COM9 );					// Close the currently opened USB comport (closing COM9, COM10 or COM11 makes no difference)
							delay(5);							// USB specification suggest we should wait 100ms when switching interface
							comopen(app.default_interface);		// Open the new USB interface
						}

						break;

					case INVALID_OPTION_READ:
						sound(TLONG, VHIGH, SLOW, 0);
						break;

					case VALID_OPTION_READ:
						sound(TSTANDARD, VHIGH, SHIGH, SMEDIUM, SHIGH,0);
						break;

					case LABEL_IGNORED:
						break;

					case ERROR:
						scannerpower(OFF, 0);
						break;
				}
			}
		}

		return MENU_CODE;
	}
	else if(code->id == MENU_CODE_PDF || code->id == MENU_CODE_C128)		// Check for Code-128 / PDF417 menu labels
	{
		switch( (result=ExecuteMenuLabel(code)) )
		{
			case INVALID_OPTION_READ:
				scannerpower(OFF, 0);
				sound(TLONG, VHIGH, SLOW, 0);
				break;

			case VALID_OPTION_READ:
				scannerpower(OFF, 0);
				systemsetting("Z2");		// Save changed settings in flash memory
				sound(TSTANDARD, VHIGH, SMEDIUM, SHIGH, SMEDIUM, SHIGH, 0);
				delay(TSTANDARD * 4);		// Wait till buzzer sound if finished before saving (saving flash settings turns off sound timer temporarily)

				if( comopen(-1) != app.default_interface )
				{
					comclose( COM9 );					// Close the currently opened USB comport (closing COM9, COM10 or COM11 makes no difference)
					delay(5);							// USB specification suggest we should wait 100ms when switching interface
					comopen(app.default_interface);		// Open the new USB interface
				}
				break;

			case LABEL_IGNORED:
			default:
				break;
		}

		return MENU_CODE;
	}

	//
	// Store barcode
	//
	if(last_key_pressed == CLR_KEY && !(app.output_quantity_options & QUANTITY_DELETE_KEY_DISABLED) )
	{
		result = StoreBarcode(code, -1);	// Substract 1 from database (if present/allowed)

		if(result >= ERROR_VALUE)
		{
			GoodReadLed(RED);
			ErrBeep();
		}
		else
		{
			if(result >= 0)		
			{
				GoodReadLed(ORANGE);	// ORANGE led indicates a succesfully deleted/substracted barcode
				OkBeep(2000);			// Barcode deleted, quantity >= 0 : 2000 Hz
			}
			else
			{
				GoodReadLed(RED);		// RED led indicates a succesfully substracted barcode, resulting in a negative value
				OkBeep(1500);			// Barcode deleted, quantity < 0 : 1500 Hz
			}
		}
	}
	else 
	{
		result = StoreBarcode(code, 1);		// Add 1 to database

		if(result >= ERROR_VALUE)
		{
			GoodReadLed(ORANGE);
			ErrBeep();
		}
		else 
		{
			if(result >= 0)
			{
				GoodReadLed( (result > 0) ? GREEN : ORANGE );
				OkBeep(3000);		// Barcode added, quantity >= 0 : 3000 Hz
			}
			else
			{
				GoodReadLed(RED);	// RED led indicates a succesfully added barcode, but still resulting in a negative value
				OkBeep(1500);		// Barcode added, quantity < 0 : 1500 Hz
			}
		}
	}

	return FALSE;
}

/**
 * Select USB VCP Interface (and RS232 when placed in cradle)
 */
void UsbSetVcp(void)
{
	Application_Default();
	app.default_interface = COM0;
}
/**
 * Select USB MSD Interface
 */
int UsbSetMsd(void)
{
	char *version = GetOsVersion();

	if( strcmp( &version[4], "0220d" ) < 0 )	// Make sure the OS supports USB-MSD (equal or higher than RBxV0220c)
		return ERROR;

	Application_Default();
	app.default_interface = COM11;
	return OK;
}

//------------------------------------------------------------------
//  main
//  =====================
//  Main function of this demo that allows the OPN200x to read
//  barcodes and transmit them using the mini USB cable to a virtual
//  comport on a PC or laptop. This application also support the
//  reading of the univeral menu-book. Since all changed
//  configurations are stored in flash, the changed configurations
//  are restored after a restart even if the battery has been empty.
//------------------------------------------------------------------
void main(void)
{
	int state, res, ch;

	//
	// Initialize the barcode structure
	//
	code.text = bcr_buf;
	code.min = 0;
	code.max = MAX_BARCODE_LENGTH;

	//
	// The following function implements systemsetting menu options that aren't present in the OS or need to be overruled
	//
	AddMenuOptionTable(menu_option_table, MAX_OPTIONS);

	//
	// Enable remembering of (decoder) settings in flash and load old settings if present
	//
	if(SystemSettingsMemorizing(ON) < 0)	// Enable remembering of (decoder) settings in flash and load old settings if present
	{
		Application_Default();				// If no old settings were found -> Reset OS and application to default
	}

	//
	// Enable remembering of application settings in flash and load old settings if present and compatible
	//
	if( ApplicationSettingsMemorizing(ON, (void *)&app, sizeof(app), VERSION_CHECK) < 0 )
	{
		Application_Default();			// If no valid application settings were found -> Reset to OS and application default
	}

	// Select the USB interface
	if(app.default_interface == COM11)
	{
		comclose(COM9);		// By default the USB-VCP port is automatically opened at start-up to allow automatic downloading, so we need to close that port first
	}

	comopen( app.default_interface );

	systemsetting("YC");				// Always makes sure menu-labels are enabled after a restart
	systemsetting("7G");				// Always makes sure charge indicator is enabled after a restart

	SetReadMode(app.read_mode, app.read_time);

	state = STATE_IDLE;					// Idle start

	autopowerdown(ON, IDLE_POWER_DOWN_TIME);	// Init auto power down time

	if(app.buzzer_options & STARTUP_BUZZER)
	{
		sound( TSTANDARD, VHIGH, SLOW, SMEDIUM, SHIGH, SLOW, SMEDIUM, SHIGH, 0 );		// Start-up beep
	}

	for(;;)
	{
		//
		// Main state machine
		//
		switch(state)
		{
			case STATE_IDLE:			// Idle state (not connected)

				if( (ch=getchar()) > 0 )
					last_key_pressed = ch;	// Used to determine whether a barcode was read using the clear key or trigger key

				//
				// Barcode reading and handling
				//
				if(readbarcode(&code) == OK)				// If barcode read
				{
					res = ProcessBarcode(&code);			// Check if a special

					if(res == MENU_CODE)					// If Menu label(s) read and executed
					{
						state = STATE_WAIT_FOR_RELEASE;		// Forces an update of the read mode, which may have been changed
						break;
					}
				}

				if(ischarging())
				{
					if(UsbStatus() != USB_NOT_CONNECTED)
					{
						if(app.default_interface == COM11)				// If USB-MSD -> Disable barcode reading, because the file system can be changed externally
							state = STATE_USB_CONNECTED;
						else if(fsize((char*)DBASE_NAME) != -1L)		// If USB-VCP and data present -> Try to transmit data using NetO
							state = STATE_USB_CONNECTED;
					}
					else if(fsize((char*)DBASE_NAME) != -1L)			// If Cradle (RS232) and data present -> Try to transmit data using NetO
						state = STATE_RS232_CONNECTED;
									
					if(state != STATE_IDLE)
					{ 
						//systemsetting("8Z");				// Disable charge indicator (or we wouldn't see the green flashing)
						//goodreadled(GREEN_FLASH, 0);    
						//systemsetting("7G");				// Re-enable charge indicator.
						scannerpower(OFF, 0);				// Disable scanning while waiting to transmit its data
					}
				}
				break;


			case STATE_USB_CONNECTED:

				if(UsbStatus() == USB_NOT_CONNECTED)
				{
					state = STATE_WAIT_FOR_RELEASE;		// Forces an update of the read mode, which may have been changed
					break;
				}
				
				if(app.default_interface != COM11)
				{
					if(TransmitData() == OK)
					{
						state = STATE_WAIT_FOR_RELEASE;		// Forces an update of the read mode, which may have been changed
					}
				}
				else
				{
					while(ischarging())					// Wait till USB cable is removed
						idle();

					state = STATE_WAIT_FOR_RELEASE;		// Forces an update of the read mode, which may have been changed
				}
				break;

			case STATE_RS232_CONNECTED:

				if(!ischarging() || UsbStatus() != USB_NOT_CONNECTED)
				{
					state = STATE_WAIT_FOR_RELEASE;		// Forces an update of the read mode, which may have been changed
					break;
				}
				else if(UsbStatus() != USB_NOT_CONNECTED)
				{
					state = STATE_USB_CONNECTED;
					break;
				}

				if(TransmitData() == OK)
				{
					state = STATE_WAIT_FOR_RELEASE;		// Forces an update of the read mode, which may have been changed
					break;
				}

				break;

			case STATE_WAIT_FOR_RELEASE:
			default:
				if(!triggerpressed() && !downpressed())
				{
					//goodreadled(GREEN_FLASH, 0);		// Stop the green flashing LED (if enabled)
					//systemsetting("7G");				// Re-enabled charge indicator

					state = STATE_IDLE;
					SetReadMode(app.read_mode, app.read_time);
				}
				break;
		}

		idle();		// This reduces the power consumption
	}
}
