//------------------------------------------------------------------
//  Bluetooth Demo application for the OPN2002
//------------------------------------------------------------------
//  This Bluetooth demo allows the OPN2002 to read barcodes and
//  transmit them by Bluetooth as either serial data to Bluetooth virtual
//  com port or as Bluetooth keyboard data.
//  (When not connected it transmits it's data by USB (VCP) or serially by
//  the CRD2000-RU10 multi cradle)
//
//  This application also supports the reading of the Univeral menu-book
//  making it possible to configure all sort of barcode and Bluetooth options.
//  Also OSE serial commands can be used to configure the various options.
//
//  Since all changed configurations are stored in flash, any changed
//  configurations are restored after a restart or when the battery was empty.
//
//  This application does not demonstrate the storing of barcode
//  data in a file or in flash memory.
//
//  See the application manual on more details on the support
//  features and on how the application works.
//
//------------------------------------------------------------------
//  Version history:
//     RFG35310		:	First preliminary release
//                      Minimal required OS version: RBGV0203c
//						Release date: October 7, 2009
//						Author: Ronny de Winter
//						Company: Opticon Sensors Europe BV
//------------------------------------------------------------------
//     RFG35311		:	Bugfixes and added following new features:
//                      - configurable auto reconnect time
//                      - configurable trigger to (dis)connect time
//                      - support of connect/disconnect labels
//                      Minimal required OS version: RBGV0204
//                      Release date: October 16, 2009
//						Author: Ronny de Winter
//						Company: Opticon Sensors Europe BV
//------------------------------------------------------------------
//     RFG35312		:	Added following new features:
//                      - configurable read mode / read time
//                      - configurable buzzer options
//                      - configurable good read LED options
//						- serial OSE command support
//						- Added a retry to a connection attempt
//						- Made a reconnect attempt cycle faster (20s -> 15s)
//                      Minimal required OS version: RBGV0208
//                      Release date: January 18, 2010
//						Author: Ronny de Winter
//						Company: Opticon Sensors Europe BV
//------------------------------------------------------------------
//     RFG35313		:	- Made connectable time configurable (CT01...CT10)
//                      - Fixed reconnect time in slave mode
//						- Manual update
//                      Minimal required OS version: RBGV0208
//                      Recommended OS version: RBGV0215a or higher
//                      Release date: July 30, 2010
//						Author: Ronny de Winter
//						Company: Opticon Sensors Europe BV
//------------------------------------------------------------------
//     RFG35314		:	- Added option to toggle iOS on screen keyboard
//						  when connected to an iPhone / iPad
//                      - Manual update
//                      Minimal required OS version: RBGV0216
//                      Recommended OS version: RBGV0219 or higher
//                      Release date: November 1, 2010
//						Author: Ronny de Winter
//						Company: Opticon Sensors Europe BV
//------------------------------------------------------------------
//     RFG35315		:	- Automatic wake-up of iPhones using the trigger key
//						- Fixed repeating character bug on iPhones/iPads
//                      - Added protection against electrical discharges (ESD)
//                      - Manual update
//                      Minimal required OS version: RBGV0219
//                      Recommended OS version: RBGV0219 or higher
//                      Release date: February 11, 2011
//						Author: Ronny de Winter
//						Company: Opticon Sensors Europe BV
//------------------------------------------------------------------
//     RFG35316		:	Added support for USB-HID (SU) and USB-VCP (C01),
//						which always barcodes to be transmitted by USB-VCP or USB-HID
//						when the OPN2002 is connected by its mini-USB cable.
//                      Minimal required OS version: RBGV0220d
//                      Recommended OS version: RBGV0221 or higher
//                      Release date: June 20, 2011
//						Author: Ronny de Winter
//						Company: Opticon Sensors Europe BV
//------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "lib.h"
#include "RFG3531.h"
#include "cmd_handler.h"

static char bcr_buf[MAX_BARCODE_LENGTH];
static struct barcode code = {0};

// Return values of ProcessBarcode
#define NORMAL_BARCODE		0				// A normal barcode label was read
#define MENU_LABEL			1				// A normal menu label was read
#define CONNECT_SAME		2				// A valid Bluetooth address label was read that matches the address that was currently stored in flash
#define CONNECT_NEW			3				// A valid Bluetooth address label was read that was different than the address that was currently stored in flash
#define DISCONNECT_LABEL	4				// A +-DISC-+ label was read
#define CONNECT_LABEL		5				// A +-CONN-+ label was read
#define CONNECTABLE_LABEL	6				// A +-DSCO-+ label was read
#define CONFIG_IPHONE_LABEL 7				// A +-IOSON-+ or +-IOSOFF-+ lable was read

// Bluetooth address / PIN code defines
#define BLT_BD_ADDR_LEN		12				// Length of a Bluetooth address (label)
#define BLT_MAX_PIN_LEN		16				// Maximum length of a Bluetooth PIN code (label)

// Defines and variables used for the various timers used by this application

#define MAX_TIMERS				8
#define CONNECT_MASTER			0		// Timer used for 'Trigger to connect' time
#define MAKE_CONNECTABLE		1		// Timer used for 'Trigger to make connectable' time
#define CONNECTING_TIME_OUT		2		// Timer used for time-out mechanism on a connection attempt
#define CONNECTABLE_TIME_OUT	3		// Timer used for time-out mechanism when connectable
#define DISCONNECT				4		// Timer used for 'Trigger to disconnect' time
#define BLUE_FLASH_SLOW		    5		// Timer used for slow blinking blue connection LED (indicating an active connection)
#define PIN_REQUEST_TIME_OUT	6		// Timer used for simulating the menu book beeping sound
#define PIN_REQUEST_MENU_BEEP	7		// Timer used for simulating the menu book beeping sound when entering a PIN code

#define TRIGGER_TO_CONNECT_TIME			  5			// Trigger to connect time: 5 seconds (using the trigger key)
#define TRIGGER_TO_CONNECTABLE_TIME		  5			// Trigger to make connectable time: 5 seconds (using the clear key)
#define TRIGGER_TO_DISCONNECT_TIME		  5			// Trigger to disconnect time: 5 seconds (using the clear key)
#define RECONNECT_WAIT_TIME			      5			// Time till between reconnects attempt: 5 seconds
#define RECONNECT_ATTEMPTS			      8			// Default number of reconnect attempts after a connection loss: 8 ( 8 x (10+5) = 120 seconds)
#define CONNECTABLE_TIME			     12			// Time-out when connectable: 120 seconds (stored in 10s steps)
#define CONNECTING_TIME				     10			// Time-out on a connecting attempt: 10 seconds
#define RECONNECTING_TIME			     10			// Time-out on a reconnecting attempt: 10 seconds

#define BLUE_FLASH_SLOW_TIME		   5*50			// Slow blinking blue connection LED: blink every 5 seconds
#define IDLE_POWER_DOWN_TIME		   3*50			// Powerdown time when idle: 3 seconds
#define CONNECTED_POWER_DOWN_TIME	  10*50			// Powerdown (sleep) time when connected: 10 seconds

#define PIN_REQUEST_TIME			   30			// Max. time allowed to enter the pin code (30 seconds)
#define PIN_REQUEST_MENU_BEEP_TIME	   1*50			// Time between beeps of the menu book beeping sound

volatile unsigned int timer[MAX_TIMERS];			// Array of various timers

//
// States of the main state machine of this application
//
#define STATE_DISCONNECTED		 0		// Idle state: No Bluetooth activity
#define STATE_CONNECTING		 1		// Connecting state: Trying to connect (as master)
#define STATE_CONNECTABLE		 2		// Connectable state: Waiting for a connection (as slave) & discoverable
#define STATE_CONNECTED_MASTER	 3		// Connected (as master) state
#define STATE_CONNECTED_SLAVE	 4		// Connected (as slave) state
#define STATE_WAIT_TO_RECONNECT	 5		// Wait to reconnect state: After a connection loss, waits until reconnect time has passed and then tries to reconnect
#define STATE_WAIT_FOR_RELEASE	 6		// Wait state: Waits until both keys are not pressed and then jumps back idle state

// Character to go to the next line depends on the interface
#define NEXT_LINE	(comopen(-1) == COM10 || comopen(-1) == COM12 || comopen(-1) == COM13) ? KB_ENT_KEY : '\r'

//
// local functions
//
static void Application_Default(void);
static void Bluetooth_VCP_Default(void);
static void Bluetooth_HID_Default(void);
static void SetUsbHidMode(void);
static void SetUsbVcpMode(void);
static void Version_Output(void);

// Global variables
application_vars app;			// Structure of configuration settings that is stored in non-volatile (flash) memory
int reset_appl_to_default;
int reset_read_mode;

//
// The following table implements systemsetting options that need to be added to the OS or overruled
//
// Note: The use of 'systemsetting()' inside the functions in the table below is limited to a protect against recursion,
//       It is still possible to use systemsetting() but only with a 1 menu option per function call.
//
const OPTION blt_menu_option_table[] =
{                                               // CLEAR mask                              SET mask (use 0xFF 0xFF in case of a function call)
	{ "GC",(void*)&(app.buzzer_options),				0x00,                                   STARTUP_BUZZER              },
	{ "GD",(void*)&(app.buzzer_options),				STARTUP_BUZZER,                         0x00                        },
	{ "S0",(void*)&(app.read_mode),						0xFF,                                   SINGLE                      },
	{ "S1",(void*)&(app.read_mode),						0xFF,                                   MULTIPLE                    },
	{ "S2",(void*)&(app.read_mode),						0xFF,                                   CONTINUOUS                  },
	{ "SO",(void*)Bluetooth_VCP_Default,				0xFF,                                   0xFF                        },
    { "SU",(void*)SetUsbHidMode,						0xFF,                                   0xFF                        },
	{ "T0",(void*)&(app.buzzer_volume),					0xFF,                                   VHIGH                       },
	{ "T1",(void*)&(app.buzzer_volume),					0xFF,                                   VMEDIUM                     },
	{ "T2",(void*)&(app.buzzer_volume),					0xFF,                                   VSTANDARD                   },
	{ "T3",(void*)&(app.buzzer_volume),					0xFF,                                   VLOW                        },
	{ "T4",(void*)&(app.goodread_duration),				0xFF,                                   0                           },
	{ "T5",(void*)&(app.goodread_duration),				0xFF,                                   10                          },	// 200ms
	{ "T6",(void*)&(app.goodread_duration),				0xFF,                                   20                          },	// 400ms
	{ "T7",(void*)&(app.goodread_duration),				0xFF,                                   40                          },	// 800ms
	{ "TS",(void*)&(app.ser_cmd_options),				0x00,                                   RS_CONFIG_DISABLED          },
	{ "TT",(void*)&(app.ser_cmd_options),				RS_CONFIG_DISABLED,                     0x00                        },
	{ "U2",(void*)Bluetooth_VCP_Default,				0xFF,                                   0xFF                        },
	{ "W0",(void*)&(app.buzzer_options),				BUZZER_ENABLED,                         0x00                        },
	{ "W1",(void*)&(app.buzzer_options),				BUZZER_TONE_MASK,                       BUZZER_TONE_SINGLE          },
	{ "W2",(void*)&(app.buzzer_options),				BUZZER_TONE_MASK,                       BUZZER_TONE_HIGH_LOW        },
	{ "W3",(void*)&(app.buzzer_options),				BUZZER_TONE_MASK,                       BUZZER_TONE_LOW_HIGH        },
	{ "W4",(void*)&(app.buzzer_duration),				0xFF,                                   TSHORT                      },  // 100ms (=actually 120ms)
	{ "W5",(void*)&(app.buzzer_duration),				0xFF,                                   TSTANDARD                   },  // 200ms
	{ "W6",(void*)&(app.buzzer_duration),				0xFF,                                   TLONG                       },  // 400ms
	{ "W7",(void*)&(app.buzzer_duration),				0xFF,                                   TCLICK                      },  // 50ms (=actually 40ms)
	{ "W8",(void*)&(app.buzzer_options),				0x00,                                   BUZZER_ENABLED              },
    { "WC",(void*)&(app.ser_cmd_options),				0x00,                                   RS_ACK_NAK_ON               },
    { "WD",(void*)&(app.ser_cmd_options), 				RS_ACK_NAK_ON,                          0x00                        },
	{ "Y0",(void*)&(app.read_time),						0xFF,                                   0                           },
	{ "Y1",(void*)&(app.read_time),						0xFF,                                   10                          },
	{ "Y2",(void*)&(app.read_time),						0xFF,                                   20                          },
	{ "Y3",(void*)&(app.read_time),						0xFF,                                   30                          },
	{ "Y4",(void*)&(app.read_time),						0xFF,                                   40                          },
	{ "Y5",(void*)&(app.read_time),						0xFF,                                   50                          },
	{ "Y6",(void*)&(app.read_time),						0xFF,                                   60                          },
	{ "Y7",(void*)&(app.read_time),						0xFF,                                   70                          },
	{ "Y8",(void*)&(app.read_time),						0xFF,                                   80                          },
	{ "Y9",(void*)&(app.read_time),						0xFF,                                   90                          },
	{ "YM",(void*)&(app.read_time),						0xFF,                                   READ_TIME_INDEFINITELY      },
	{ "Z1",(void*)Version_Output,						0xFF,                                   0xFF                        },
    { "C01",(void*)SetUsbVcpMode,						0xFF,                                   0xFF                        },
	{ "C02",(void*)Bluetooth_HID_Default,				0xFF,                                   0xFF                        },
	{ "C05",(void*)Bluetooth_VCP_Default,				0xFF,                                   0xFF                        },
	{ "XIR",(void*)&(app.ser_cmd_format),				CMD_TERM_MASK,                          CMD_TERM_CR|CMD_TERM_ETX    },
	{ "XIS",(void*)&(app.ser_cmd_format),				CMD_TERM_MASK,                          CMD_TERM_ETX                },
	{ "XIT",(void*)&(app.ser_cmd_format),				CMD_TERM_MASK,                          CMD_TERM_CR                 },
	{ "XIU",(void*)&(app.ser_cmd_format),				CMD_TERM_MASK,                          CMD_TERM_LF                 },
	{ "XIV",(void*)&(app.ser_cmd_format),				CMD_TERM_MASK,                          CMD_TERM_CR|CMD_TERM_LF     },
	{ "XIW",(void*)&(app.ser_cmd_format),				CMD_TERM_MASK,                          CMD_TERM_EOT                },
	{ "XIX",(void*)&(app.ser_cmd_format),				CMD_HEADER_MASK,                        CMD_HEADER_ESC|CMD_HEADER_STX},
	{ "XIY",(void*)&(app.ser_cmd_format),				CMD_HEADER_MASK,                        CMD_HEADER_STX              },
	{ "XIZ",(void*)&(app.ser_cmd_format),				CMD_HEADER_MASK,                        0                           },
	{ "FPIN",(void*)&(app.blt_hid_manual_pin),			0x01,                                   0x00                        },
	{ "CA00",(void*)&(app.reconnect_attempts),			0xFF,                                   0                           },
	{ "CA01",(void*)&(app.reconnect_attempts),			0xFF,                                   4                           },
	{ "CA02",(void*)&(app.reconnect_attempts),			0xFF,                                   8                           },
	{ "CA03",(void*)&(app.reconnect_attempts),			0xFF,                                   12                          },
	{ "CA04",(void*)&(app.reconnect_attempts),			0xFF,                                   16                          },
	{ "CA05",(void*)&(app.reconnect_attempts),			0xFF,                                   20                          },
	{ "CA06",(void*)&(app.reconnect_attempts),			0xFF,                                   24                          },
	{ "CA07",(void*)&(app.reconnect_attempts),			0xFF,                                   28                          },
	{ "CA08",(void*)&(app.reconnect_attempts),			0xFF,                                   32                          },
	{ "CA09",(void*)&(app.reconnect_attempts),			0xFF,                                   36                          },
	{ "CT01",(void*)&(app.connectable_time),			0xFF,                                   6                           },
	{ "CT02",(void*)&(app.connectable_time),			0xFF,                                   12                          },
	{ "CT03",(void*)&(app.connectable_time),			0xFF,                                   18                          },
	{ "CT04",(void*)&(app.connectable_time),			0xFF,                                   24                          },
	{ "CT05",(void*)&(app.connectable_time),			0xFF,                                   30                          },
	{ "CT06",(void*)&(app.connectable_time),			0xFF,                                   36                          },
	{ "CT07",(void*)&(app.connectable_time),			0xFF,                                   42                          },
	{ "CT08",(void*)&(app.connectable_time),			0xFF,                                   48                          },
	{ "CT09",(void*)&(app.connectable_time),			0xFF,                                   54                          },
	{ "CT10",(void*)&(app.connectable_time),			0xFF,                                   60                          },
	{ "PC00",(void*)&(app.trigger_to_connect_time),		0xFF,                                   0x00                        },
	{ "PC01",(void*)&(app.trigger_to_connect_time),		0xFF,                                   0x01                        },
	{ "PC02",(void*)&(app.trigger_to_connect_time),		0xFF,                                   0x02                        },
	{ "PC03",(void*)&(app.trigger_to_connect_time),		0xFF,                                   0x03                        },
	{ "PC04",(void*)&(app.trigger_to_connect_time),		0xFF,                                   0x04                        },
	{ "PC05",(void*)&(app.trigger_to_connect_time),		0xFF,                                   0x05                        },
	{ "PC06",(void*)&(app.trigger_to_connect_time),		0xFF,                                   0x06                        },
	{ "PC07",(void*)&(app.trigger_to_connect_time),		0xFF,                                   0x07                        },
	{ "PC08",(void*)&(app.trigger_to_connect_time),		0xFF,                                   0x08                        },
	{ "PC09",(void*)&(app.trigger_to_connect_time),		0xFF,                                   0x09                        },
	{ "PD00",(void*)&(app.trigger_to_disconnect_time),  0xFF,                                   0x00                        },
	{ "PD01",(void*)&(app.trigger_to_disconnect_time),	0xFF,                                   0x01                        },
	{ "PD02",(void*)&(app.trigger_to_disconnect_time),	0xFF,                                   0x02                        },
	{ "PD03",(void*)&(app.trigger_to_disconnect_time),	0xFF,                                   0x03                        },
	{ "PD04",(void*)&(app.trigger_to_disconnect_time),	0xFF,                                   0x04                        },
	{ "PD05",(void*)&(app.trigger_to_disconnect_time),	0xFF,                                   0x05                        },
	{ "PD06",(void*)&(app.trigger_to_disconnect_time),	0xFF,                                   0x06                        },
	{ "PD07",(void*)&(app.trigger_to_disconnect_time),	0xFF,                                   0x07                        },
	{ "PD08",(void*)&(app.trigger_to_disconnect_time),	0xFF,                                   0x08                        },
	{ "PD09",(void*)&(app.trigger_to_disconnect_time),	0xFF,                                   0x09                        },
	{ "MPIN",(void*)&(app.blt_hid_manual_pin),			0x00,                                   0x01                        },
	{ "UNPR",(void*)blt_unpair,							0xFF,                                   0xFF                        },
};

#define MAX_OPTIONS sizeof(blt_menu_option_table)/sizeof(OPTION)

//------------------------------------------------------------------
//  Application_Default
//  ============================
//  Reset all application specific settings to default
//------------------------------------------------------------------
void Application_Default(void)
{
	reset_appl_to_default = FALSE;

	app.blt_hid_manual_pin = TRUE;									// PIN code for HID must be entered using numeric direct input labels (Menu book chapter 4.3.3)
	app.trigger_to_connect_time = TRIGGER_TO_CONNECT_TIME;			// Trigger to connect time (5 seconds)
	app.trigger_to_connectable_time = TRIGGER_TO_CONNECTABLE_TIME;	// Trigger to make connectable time (5 seconds)
	app.trigger_to_disconnect_time = TRIGGER_TO_DISCONNECT_TIME;	// Trigger to disconnect time (5 seconds)
	app.reconnect_attempts = RECONNECT_ATTEMPTS;					// Default number of reconnect attempts after connection loss (4)
	app.connectable_time = CONNECTABLE_TIME;						// Time that the device stays connectable (120 seconds)

	app.read_mode = SINGLE;											// Single read
	app.read_time = 0;												// 2 seconds
	app.goodread_duration = 10;										// Good read led duration (200ms)
	app.buzzer_volume = VHIGH;										// Buzzer volume (high)
	app.buzzer_duration = TSTANDARD;								// Buzzer duration (200ms)
	app.buzzer_options = BUZZER_ENABLED | STARTUP_BUZZER;			// Buzzer enabled, startup buzzer enabled
	app.buzzer_options |= BUZZER_TONE_HIGH_LOW;						// Buzzer tone (High-low tone)
	app.ser_cmd_options = 0;										// Serial command options (OSE serial commands allowed)
	app.ser_cmd_format = CMD_HEADER_ESC | CMD_TERM_CR;				// Supported serial command format: ESC / CR
	app.ser_cmd_format |= CMD_HEADER_STX | CMD_TERM_ETX;			// Supported serial command format: STX / ETX

	if(app.blt_interface == BLT_HID_DEVICE)
	{
		systemsetting("RZ7I");								// Set suffix: Enter key
	}
	else
	{
		app.blt_interface = BLT_VCP_DEVICE;
		systemsetting("RZ1M");								// Set suffix: <CR>
	}

	blt_unpair();
}

//------------------------------------------------------------------
//  Bluetooth_VCP_Default
//  ============================
//  Reset all Bluetooth settings to default
//------------------------------------------------------------------
void Bluetooth_VCP_Default(void)
{
	systemsetting("U2");			// Call OS default function
	systemsetting("YC");			// Enable menu-labels (as data)

	app.blt_interface = BLT_VCP_DEVICE;						// Select interface: VCP
	
	app.options &= ~(USB_VCP_MODE | USB_HID_MODE);			// Deselect USB-VCP and USB-HID transmission

	//
	// Note: A full reset is deliberatedly not executed in menu-option functions Bluetooth_VCP_Default / Bluetooth_HID_Default, so please don't change this!
	//
	// This is done because the possibity to use 'systemsetting()' in a function that's indirectly called by 'systemsetting() itself is limited due to
	// a protection against recursion. Therefor it is not possible to configure complex menu option strings, i.e. pre- and suffixes.
	//
	reset_appl_to_default = TRUE;
}

//------------------------------------------------------------------
//  Bluetooth_HID_Default
//  ============================
//  Reset all Bluetooth settings to default
//------------------------------------------------------------------
void Bluetooth_HID_Default(void)
{
	systemsetting("U2");			// Call OS default function
	systemsetting("YC");			// Enable menu-labels (as data)

	app.blt_interface = BLT_HID_DEVICE;						// Select interface

	app.options &= ~(USB_VCP_MODE | USB_HID_MODE);			// Deselect USB-VCP and USB-HID transmission

	reset_appl_to_default = TRUE;
}

//------------------------------------------------------------------
//  SetUsbHidMode
//  ============================
//  When connected to USB -> use USB-HID profile
//------------------------------------------------------------------
void SetUsbHidMode(void)
{
	char *version = GetOsVersion();

	Bluetooth_HID_Default();

	if( strcmp( &version[4], "0220d" ) >= 0 )	// Make sure the OS supports USB-MSD (equal or higher than RBxV0220d)
	{
		app.options |= USB_HID_MODE;		// Enable transmission of barcode data through USB-HID when not connected to Bluetooth
	}
}

//------------------------------------------------------------------
//  SetUsbHidMode
//  ============================
//  When connected to USB -> use USB-HID profile
//------------------------------------------------------------------
void SetUsbVcpMode(void)
{
	Bluetooth_VCP_Default();
	app.options |= USB_VCP_MODE;		// Enable transmission of barcode data through USB-HID when not connected to Bluetooth
}

//------------------------------------------------------------------
//  Version_Output
//  ============================
//  Output all versions, including the application version
//------------------------------------------------------------------
void Version_Output(void)
{
	systemsetting("Z1");

	printf("%s%c", VERSION, NEXT_LINE);
}

//------------------------------------------------------------------
//  SetReadMode
//  ============================
//  Converts the configured read time/mode to a scannerpower call
//------------------------------------------------------------------
void SetReadMode(int readmode, int readtime)
{
	switch(readmode)
	{
		case MULTIPLE:
			if(readtime == 0)
				scannerpower(MULTIPLE | TRIGGER, 10*50);		// Limit trigger mode to 10 seconds to avoid pressing the trigger for 20s, causing an unintended restart
			else if(readtime == READ_TIME_INDEFINITELY)
				scannerpower(MULTIPLE, -1);
			else
				scannerpower(MULTIPLE, readtime*5);
			break;

		case CONTINUOUS:
			if(readtime == 0)
				scannerpower(CONTINUOUS | TRIGGER, 10*50);		// Limit trigger mode to 10 seconds to avoid pressing the trigger for 20s, causing an unintended restart
			else if(readtime == READ_TIME_INDEFINITELY)
				scannerpower(CONTINUOUS, -1);
			else
				scannerpower(CONTINUOUS, readtime*5);
			break;

		case SINGLE:
		default:
			if(readtime == 0)
				scannerpower(SINGLE | TRIGGER, 10*50);			// Limit trigger mode to 10 seconds to avoid pressing the trigger for 20s, causing an unintended restart
			else if(readtime == READ_TIME_INDEFINITELY)
				scannerpower(SINGLE, -1);
			else
				scannerpower(SINGLE, readtime*5);
			break;
	}
}

//------------------------------------------------------------------
//  OkBeep
//  ============================
//  Sounds a good read beep according to the current configuration
//------------------------------------------------------------------
void OkBeep(int base_frequency)
{
	unsigned char length, volume;

	if(!(app.buzzer_options & BUZZER_ENABLED))
		return;

	volume = app.buzzer_volume;
	length = app.buzzer_duration;

	switch( app.buzzer_options & BUZZER_TONE_MASK )
	{
		case BUZZER_TONE_SINGLE:
			sound( length * 2, volume, base_frequency, 0);
			break;

		case BUZZER_TONE_LOW_HIGH:
			sound( length, volume, base_frequency, base_frequency+1000, 0);
			break;

		case BUZZER_TONE_HIGH_LOW:
		default:
			sound( length, volume, base_frequency, base_frequency-500, 0);
			break;
	}
}

//------------------------------------------------------------------
//  ErrBeep
//  ============================
//  Sounds an error beep according to the current configuration
//------------------------------------------------------------------
void ErrBeep(void)
{
	unsigned char length, volume;

	volume = app.buzzer_volume;
	length = app.buzzer_duration;

	sound( length * 2, volume, SLOW, 0);
}

//------------------------------------------------------------------
//  GoodReadLed
//  ============================
//  Shows a good read LED according to the current configuration
//------------------------------------------------------------------
void GoodReadLed(int color)
{
	goodreadled(color, app.goodread_duration);
}

//------------------------------------------------------------------
//  BluetoothTimers_Heartbeat
//  ============================
//  Heartbeat routine that is executed by the OS every 20ms, causing
//  the seperate Bluetooth timers to be decremented (if enabled)
//------------------------------------------------------------------
void BluetoothTimers_Heartbeat( void )
{
	for( int x = 0; x < MAX_TIMERS; x++ )
	{
		if( timer[x] != 0 && timer[x] != 0xFFFF )
			timer[x]--;
	}
}

//------------------------------------------------------------------
//  ProcessBarcode
//  =====================
//  Processes a barcode that has been read. This function
//  makes is possible to read set-up labels from the universal
//  book as well. (Changed settings will be lost after a restart)
//------------------------------------------------------------------
int ProcessBarcode(struct barcode *code)
{
	int i, result = 0;
	char ch;

	if(code->id == MENU_CODE || code->id == MENU_CODE_PDF || code->id == MENU_CODE_C128)	// If a menu label -> check for Bluetooth address, pin and other configuration labels
	{
		if(code->length >= 1 && code->length <= BLT_MAX_PIN_LEN)
		{
			if(code->text[0] == 'p')														// PIN code labels are Code-128/PDF417 menu labels starting with a 'p'
			{
				for(i=1; i<code->length; ++i)
				{
					ch = code->text[i];

					if(ch >= ' ' && ch <= '~')
						continue;

					break;
				}

				if(i == code->length)														// If valid PIN code label store it in flash and sound OK tune
				{
					SetBltPin( &code->text[1] );
					sound(TSTANDARD, VHIGH, SHIGH, SMEDIUM, SHIGH, SMEDIUM, SHIGH, SMEDIUM, SHIGH,0);
					return MENU_LABEL;
				}
			}
		}

		if(code->length == BLT_BD_ADDR_LEN)													// Bluetooth address labels are menu labels of 16 (hexidecimal) characters
		{
			for(i=0; i<BLT_BD_ADDR_LEN; ++i)
			{
				if( (code->text[i] < '0' || code->text[i] > '9') && (code->text[i] < 'A' || code->text[i] > 'F') && (code->text[i] < 'a' || code->text[i] > 'f') )
					break;
			}

			if(i == BLT_BD_ADDR_LEN)														// If valid Bluetooth address label
			{
				for(i=0; i<BLT_BD_ADDR_LEN; ++i)
				{
					if(code->text[i] >= 'a' && code->text[i] <= 'f')						// Convert to upper case characters
						code->text[i] = code->text[i] - 'a' + 'A';
				}

				if( memcmp(code->text, GetBltAddress(), BLT_BD_ADDR_LEN) != 0 )				// Compare with currently stored Bluetooth address
				{
					sound(TSTANDARD, VHIGH, SHIGH, SMEDIUM, SHIGH, SMEDIUM, SHIGH, SMEDIUM, SHIGH,0);
					delay(TSTANDARD * 8);		// Wait till buzzer sound if finished before saving (saving flash settings turns off sound timer temporarily)
					SetBltAddress(code->text);												// Store Bluetooth address and return 'CONNECT_NEW'
					return CONNECT_NEW;
				}

				sound(TSTANDARD, VHIGH, SHIGH, SMEDIUM, SHIGH, SMEDIUM, SHIGH, SMEDIUM, SHIGH,0);
				return CONNECT_SAME;														// Same Bluetooth address -> return 'CONNECT_SAME'
			}
		}

		if(strncmp(code->text, "+-DISC-+", 8) == 0)
			return DISCONNECT_LABEL;

		if(strncmp(code->text, "+-CONN-+", 8) == 0 && memcmp(GetBltAddress(), "000000000000", BLT_BD_ADDR_LEN) != 0)
			return CONNECT_LABEL;

		if(strncmp(code->text, "+-DSCO-+", 8) == 0 && blt_is_connected() != TRUE)
			return CONNECTABLE_LABEL;
		
		if( (strncmp(code->text, "+-IOSON-+", 9) == 0)  || (strncmp(code->text, "+-IOSOFF-+", 10) == 0) )
			return CONFIG_IPHONE_LABEL;
					
	}

	if(code->id == MENU_CODE)		// Check for standard Code-39 menu labels (from the OSE Universal menu book)
	{
		scannerpower(ON, -1);		// Turn laser on until menu mode is exited

		while(result != EXITING_MENU_MODE && result != ERROR && !scanneroff())
		{
			if(result == 0 || readbarcode(code) == OK)
			{
				switch( (result=ExecuteMenuLabel(code)) )
				{
					case ENTERING_MENU_MODE:
						sound(TSTANDARD, VHIGH, SHIGH, SMEDIUM, SHIGH, 0);
						break;

					case EXITING_MENU_MODE:
						scannerpower(OFF, 0);
						sound(TSTANDARD,VHIGH, SHIGH, SMEDIUM, SHIGH, 0);
						delay(TSTANDARD * 4);		// Wait till buzzer sound if finished before saving (saving flash settings turns off sound timer temporarily)
						systemsetting("Z2");		// Save changed settings in flash memory
						break;

					case INVALID_OPTION_READ:
						sound(TLONG, VHIGH, SLOW, 0);
						break;

					case VALID_OPTION_READ:
						sound(TSTANDARD, VHIGH, SHIGH, SMEDIUM, SHIGH,0);
						break;

					case LABEL_IGNORED:
						break;

					case ERROR:
						scannerpower(OFF, 0);
						break;
				}
			}
		}

		return MENU_LABEL;
	}
	else if(code->id == MENU_CODE_PDF || code->id == MENU_CODE_C128)		// Check for Code-128 / PDF417 menu labels
	{
		switch( (result=ExecuteMenuLabel(code)) )
		{
			case INVALID_OPTION_READ:
				scannerpower(OFF, 0);
				sound(TLONG, VHIGH, SLOW, 0);
				break;

			case VALID_OPTION_READ:
				scannerpower(OFF, 0);
				sound(TSTANDARD, VHIGH, SHIGH, SMEDIUM, SHIGH, SMEDIUM, SHIGH, SMEDIUM, SHIGH,0);
				delay(TSTANDARD * 8);		// Wait till buzzer sound if finished before saving (saving flash settings turns off sound timer temporarily)
				systemsetting("Z2");		// Save changed settings in flash memory
				break;

			case LABEL_IGNORED:
			default:
				break;
		}

		return MENU_LABEL;
	}

	return NORMAL_BARCODE;
}

//------------------------------------------------------------------
//  SetPinCallback
//  =====================
//  Callback function that is called by the OS when the
//  remote device request the entering of a PIN code.
//  This function makes it possible to enter the PIN code
//  using the numberic direct input character (Q0..Q9) of the
//  OSE Universal menu book.
//------------------------------------------------------------------
void SetPinCallback( char *pin, int* len )
{
	char data[2+1];
	struct barcode direct_input_label = {0};
	char new_pin[BLT_MAX_PIN_LEN+1];
	int x;

	if(app.blt_interface != BLT_HID_DEVICE || app.blt_hid_manual_pin != TRUE)
	{
		strcpy(pin, (char *)GetBltPin());
		*len = strlen(pin);
		return;
	}

	timer[PIN_REQUEST_TIME_OUT] = PIN_REQUEST_TIME * 50;				// Set time out
	timer[PIN_REQUEST_MENU_BEEP] = 0xFFFF;

	direct_input_label.min = direct_input_label.max = 2;		// Fixed length of 2
	direct_input_label.max = 16;		// Fixed length of 2
	direct_input_label.text = data;

	scannerpower(ON | MULTIPLE, -1);

	x=0;

	for(;;)
	{
		if( downpressed() )		// Press down key when done (or read SET/END label)
			break;

		if( timer[ PIN_REQUEST_TIME_OUT ] == 0 )
			break;

		if(x >= BLT_MAX_PIN_LEN)
			break;

		// Simulate the menu book beeping sound
		if( timer[PIN_REQUEST_MENU_BEEP] == 0 || timer[PIN_REQUEST_MENU_BEEP] == 0xFFFF )
		{
			 sound(TSTANDARD, VSTANDARD, SMEDIUM, 0);
			 timer[PIN_REQUEST_MENU_BEEP] = PIN_REQUEST_MENU_BEEP_TIME;
		}

		if( readbarcode(&direct_input_label) == OK )
		{
			timer[PIN_REQUEST_MENU_BEEP] = PIN_REQUEST_MENU_BEEP_TIME;

			if( strcmp( "ZZ", data ) == 0 )
			{
				sound(TSTANDARD,VHIGH, SHIGH, SMEDIUM, SHIGH, 0);
				break;
			}
			else if(data[0] == 'Q' && isdigit(data[1]))
			{
				sound(TSTANDARD, VHIGH, SHIGH, SMEDIUM, SHIGH,0);
				new_pin[x++] = data[1];
			}
			else if(data[0] == '0' && (isalpha(data[1]) || isdigit(data[1])))
			{
				sound(TSTANDARD, VHIGH, SHIGH, SMEDIUM, SHIGH,0);
				new_pin[x++] = data[1];
			}
			else if(data[0] == '$' && (isalpha(data[1]) || isdigit(data[1])))
			{
				sound(TSTANDARD, VHIGH, SHIGH, SMEDIUM, SHIGH,0);
				new_pin[x++] = data[1] - 'A' + 'a';
			}
			else
			{
				sound(TLONG, VHIGH, SLOW, 0);
			}
		}
	}

	scannerpower(OFF, 0);

	SetReadMode(app.read_mode, app.read_time);

	timer[CONNECTABLE_TIME_OUT] = app.connectable_time * 10 * 50;

	new_pin[x] = 0;
	strcpy( pin, new_pin );
	*len = strlen( pin );
}



//------------------------------------------------------------------
//  main
//  =====================
//  Main function of this demo that allows the OPN200x to read
//  barcodes and transmit them using the mini USB cable to a virtual
//  comport on a PC or laptop. This application also support the
//  reading of the univeral menu-book. Since all changed
//  configurations are stored in flash, the changed configurations
//  are restored after a restart even if the battery has been empty.
//------------------------------------------------------------------
void main(void)
{
	int state, res, ch;
	int x, timer_no, com_opened, attempts_left;
	int pressed = FALSE;

	//
	// Initialize the barcode structure
	//
	code.text = bcr_buf;
	code.min = 0;
	code.max = MAX_BARCODE_LENGTH;

	reset_appl_to_default = FALSE;
	reset_read_mode = FALSE;

	attempts_left = 0;

	//
	// The following function implements systemsetting menu options that aren't present in the OS or need to be overruled
	//
	AddMenuOptionTable(blt_menu_option_table, MAX_OPTIONS);

	//
	// Enable remembering of (decoder) settings in flash and load old settings if present
	//
	if(SystemSettingsMemorizing(ON) < 0)	// Enable remembering of (decoder) settings in flash and load old settings if present
	{
		Bluetooth_VCP_Default();			// If no old settings were found -> Reset OS and application to default
		Application_Default();
	}

	//
	// Enable remembering of application settings in flash and load old settings if present and compatible
	//
	if( ApplicationSettingsMemorizing(ON, (void *)&app, sizeof(app), VERSION_CHECK) < 0 )
	{
		Bluetooth_VCP_Default();		// If no valid application settings were found -> Reset to OS and application default
		Application_Default();
	}

	systemsetting("YC");				// Always makes sure menu-labels are enabled after a restart

	//
	// Callback function that is called by the OS when the
	// remote device request the entering of a PIN code
	//
	blt_set_pin_callback(SetPinCallback);


	for(x=0; x<MAX_TIMERS; ++x)				// Disable all timers
		timer[x] = 0xFFFF;

	//
	// Install our own heartbeat timer that beats every 20ms
	//
	InstallHeartbeatHandler( BluetoothTimers_Heartbeat );

	SetReadMode(app.read_mode, app.read_time);

	state = STATE_DISCONNECTED;					// Idle start

	autopowerdown(ON, IDLE_POWER_DOWN_TIME);	// Init auto power down time

	com_opened = -1;

	// Select the USB interface
	if(app.options & USB_HID_MODE)
	{
		comclose(COM9);		// By default the USB-VCP port is automatically opened at start-up to allow automatic downloading, so we need to close that port first
		comopen(COM10);
	}

	if(app.buzzer_options & STARTUP_BUZZER)
	{
		sound( TSTANDARD, VHIGH, SLOW, SMEDIUM, SHIGH, SLOW, SMEDIUM, SHIGH, 0 );		// Start-up beep
	}

	for(;;)
	{
		//
		// Barcode reading and handling
		//
		if(readbarcode(&code) == OK)				// If barcode read
		{
			res = ProcessBarcode(&code);			// Check if it's a special barcode that requires action

			if(res != NORMAL_BARCODE)
			{
				scannerpower(OFF, 0);				// Turn of the laser
				reset_read_mode = TRUE;				// Don't actually reset the read mode until the trigger is released
			}

			switch(res)
			{
				case NORMAL_BARCODE:
					//
					// Output barcode
					//
					if(blt_is_connected() == TRUE)
					{
						if(putstring(code.text) == OK)
						{
							GoodReadLed(GREEN);
							OkBeep(3000);		// Default frequency is 3000 Hz
						}
						else
						{
							GoodReadLed(ORANGE);
							ErrBeep();
						}
					}
					else
					{
						putstring(code.text);

						//
						// Check if USB-VCP (C01) or USB-HID (SU) default has been read -> if so, also sound OK-beep when barcode was transmitted by USB
						//
						if( ischarging() && UsbStatus() != USB_NOT_CONNECTED && ((app.options & USB_HID_MODE) || (app.options & USB_VCP_MODE)) )
						{
							GoodReadLed(GREEN);
							OkBeep(3000);		// Default frequency is 3000 Hz
						}
						else
						{
							GoodReadLed(ORANGE);
							ErrBeep();
						}
					}
					break;

				case MENU_LABEL:					// If Menu label(s) read and executed

					//
					// Check if settings need to be reset to default
					//
					// Note: A reset is deliberatedly not executed in menu-option functions Bluetooth_VCP_Default / Bluetooth_HID_Default, so please don't change this
					//
					// This is done because the possibity to use 'systemsetting()' in a function that's indirectly called by 'systemsetting() limited due to a protection against recursion
					//
					if(reset_appl_to_default)
					{
						Application_Default();
						systemsetting("Z2");

						comclose( COM9 );					// Close the currently opened USB comport (closing COM9, COM10 or COM11 makes no difference)
						delay(5);							// USB specification suggest we should wait 100ms when switching interface
						com_opened = (app.options & USB_HID_MODE) ? COM10 : COM0;		
						comopen( com_opened );				// Select either USB-HID or USB-VCP
					}

					while(triggerpressed())						// Wait for release before resetting read mode
						idle();

					SetReadMode(app.read_mode, app.read_time);
					break;

				case CONNECT_NEW:								// If different bluetooth address label read

					if(blt_is_connected() == TRUE)				// If already connected, disconnect the old, then try to connect with new address
					{
						timer[DISCONNECT] = 0xFFFF;
						goodreadled(BLUE_FLASH, 0);
						comclose(com_opened);
						com_opened = -1;
					}

					// Fall through !!!

				case CONNECT_SAME:								// If current bluetooth address read
				case CONNECT_LABEL:
				case CONNECTABLE_LABEL:

					if(blt_is_connected() == TRUE)
						break;

					timer[CONNECT_MASTER] = 0xFFFF;
					sound(TSTANDARD, VHIGH, SHIGH, SMEDIUM, 0);

					goodreadled(BLUE_FLASH, FOREVER);

					autopowerdown(OFF, 0);						// Don't turn off while connecting

					com_opened = -1;

					if(app.blt_interface == BLT_HID_DEVICE && (res == CONNECT_NEW || res == CONNECTABLE_LABEL))		// A Bluetooth HID device can only connect as slave without pair
					{
						if(comopen(COM12) == OK)				// Bluetooth HID: make connectabe
						{
							com_opened = COM12;
							state = STATE_CONNECTABLE;
							timer[CONNECTABLE_TIME_OUT] = app.connectable_time * 10 * 50;
						}
						else
						{
							comclose(COM12);
						}
					}
					else
					{
						if(app.blt_interface == BLT_VCP_DEVICE && res == CONNECTABLE_LABEL)
						{
							if(comopen(COM5) == OK)
							{
								com_opened = COM5;
								state = STATE_CONNECTABLE;
								timer[CONNECTABLE_TIME_OUT] = app.connectable_time * 10 * 50;
							}
							else
							{
								comclose(COM5);
							}
						}
						else
						{
							com_opened = (app.blt_interface == BLT_HID_DEVICE) ? COM13 : COM3;	// Connect as master as either VCP or HID

							if(comopen(com_opened) == OK)
							{
								state = STATE_CONNECTING;
								timer[CONNECTING_TIME_OUT] = CONNECTING_TIME * 50;
								attempts_left = 1;			// <- causes a 2nd connect attempt if the first failed
							}
							else
							{
								comclose(com_opened);
								com_opened = -1;
							}
						}
					}

					if(com_opened == -1)
					{
						goodreadled(BLUE_FLASH, 0);
						sound(TVLONG, VHIGH, SLOW, 0);
						state = STATE_WAIT_FOR_RELEASE;
					}
					break;

				case DISCONNECT_LABEL:

					if(state == STATE_DISCONNECTED && blt_is_connected() != TRUE)
						break;

					timer[DISCONNECT] = 0xFFFF;
					sound( TVLONG, VHIGH, SHIGH, SLOW, 0 );
					goodreadled(BLUE_FLASH, 0);
					comclose(com_opened);
					com_opened = -1;
					state = STATE_WAIT_FOR_RELEASE;
					break;

				case CONFIG_IPHONE_LABEL:			// +-IOSON-+  , +-IOSOFF-+
					goodreadled( GREEN, 10 );
					
					if( strncmp( &code.text[5], "ON" , 2 ) == 0 )
					{
						app.options |= ENABLE_KEYBOARD_TOGGLE | ENABLE_WAKEUP_KEY;
						sound(TSTANDARD, VHIGH, SMEDIUM, SHIGH,  0);
					}
					else
					{
						app.options &= ~(ENABLE_KEYBOARD_TOGGLE | ENABLE_WAKEUP_KEY);
						sound(TSTANDARD, VHIGH, SHIGH, SMEDIUM, 0);
					}
					break;
			}
		}

		//
		// Check for serial commands that may have been received and execute them
		//
		if((ch=getcom(0)) >= 0)
		{
			CheckSerialCommands(ch);

			reset_read_mode = TRUE;
			scannerpower(OFF, 0);		// Turn off the laser if the read mode has been changed
		}								// The read mode itself isn't updated before the trigger is released
		
		//
		// Check if the read mode needs to be updated
		//
		if( reset_read_mode && !triggerpressed() )
		{
			reset_read_mode = FALSE;
			SetReadMode(app.read_mode, app.read_time);
		}


		// Main Bluetooth connection state machine
		//
		switch(state)
		{
			case STATE_DISCONNECTED:			// Idle state (not connected)

				if(triggerpressed())			// Check trigger key for 'Trigger to connect' (as master)
				{
					timer[MAKE_CONNECTABLE] = 0xFFFF;

					if(timer[CONNECT_MASTER] == 0xFFFF && app.trigger_to_connect_time != 0)
					{
						timer[CONNECT_MASTER] = app.trigger_to_connect_time * 50;
					}
					else if(timer[CONNECT_MASTER] == 0)			// If 'Trigger to connect' time passed
					{
						timer[CONNECT_MASTER] = 0xFFFF;

						if(memcmp(GetBltAddress(), "000000000000", BLT_BD_ADDR_LEN) != 0)	// If no bluetooth address is configured, there's no point in trying to connect
						{
							sound(TSTANDARD, VHIGH, SHIGH, SMEDIUM, 0);

							goodreadled(BLUE_FLASH, FOREVER);

							autopowerdown(OFF, 0);				// Don't turn off while connecting

							com_opened = (app.blt_interface == BLT_HID_DEVICE) ? COM13 : COM3;	// Connect as master as either VCP or HID

							if(comopen(com_opened) == OK)		// Start the connection attempt
							{
								state = STATE_CONNECTING;
								timer[CONNECTING_TIME_OUT] = CONNECTING_TIME * 50;
								attempts_left = 1;				// <- causes a 2nd connect attempt if the first failed
							}
							else								// If Bluetooth initialization failed -> abort
							{
								goodreadled(BLUE_FLASH, 0);
								comclose(com_opened);
								com_opened = -1;
								sound(TVLONG, VHIGH, SLOW, 0);
								state = STATE_WAIT_FOR_RELEASE;
							}
						}
						else
						{
							state = STATE_WAIT_FOR_RELEASE;
						}
					}
				}
				else if(downpressed())		// Check clear key for 'Trigger to make connectable' (as slave)
				{
					timer[CONNECT_MASTER] = 0xFFFF;

					if(timer[MAKE_CONNECTABLE] == 0xFFFF && app.trigger_to_connectable_time != 0)
					{
						timer[MAKE_CONNECTABLE] = app.trigger_to_connectable_time * 50;
					}
					else if(timer[MAKE_CONNECTABLE] == 0)		// If 'Trigger to make connectable' time passed
					{
						timer[MAKE_CONNECTABLE] = 0xFFFF;

						sound( TSTANDARD, VHIGH, SHIGH, SMEDIUM, 0);

						goodreadled(BLUE_FLASH, FOREVER);

						autopowerdown(OFF, 0);		// Don't turn off while waiting for a connection

						com_opened = (app.blt_interface == BLT_HID_DEVICE) ? COM12 : COM5;	// Make connectable as slave for either VCP or HID

						if(comopen(com_opened) == OK)		// Make device connectable
						{
							state = STATE_CONNECTABLE;
							timer[CONNECTABLE_TIME_OUT] = app.connectable_time * 10 * 50;
						}
						else								// If Bluetooth initialization failed -> abort
						{
							goodreadled(BLUE_FLASH, 0);
							comclose(com_opened);
							com_opened = -1;
							sound(TVLONG, VHIGH, SLOW, 0);
							state = STATE_WAIT_FOR_RELEASE;
						}
					}
				}
				else
				{
					timer[MAKE_CONNECTABLE] = 0xFFFF;
					timer[CONNECT_MASTER] = 0xFFFF;
				}
				break;

			case STATE_CONNECTING:							// Wait for a connection to be established
			case STATE_CONNECTABLE:

				if(blt_is_connected() == TRUE)				// If Connection established -> go to connected state
				{
					sound(TVLONG, app.buzzer_volume, SHIGH, 0);
					goodreadled(BLUE, 20);
					timer[BLUE_FLASH_SLOW] = BLUE_FLASH_SLOW_TIME;
					state = (state == STATE_CONNECTING) ? STATE_CONNECTED_MASTER : STATE_CONNECTED_SLAVE;
					attempts_left = 0;
					pressed = FALSE; // no need to send a wake-up key when you've just connected (device is on already)
					autopowerdown(ON, CONNECTED_POWER_DOWN_TIME);
					break;
				}

				timer_no = (state == STATE_CONNECTING) ? CONNECTING_TIME_OUT : CONNECTABLE_TIME_OUT;

				if(downpressed())					// Clear key can be used to abort the connection attempt
				{
					if(timer[DISCONNECT] == 0xFFFF && app.trigger_to_disconnect_time != 0)
					{
						timer[DISCONNECT] = app.trigger_to_disconnect_time * 50;
					}
					else if(timer[DISCONNECT] == 0)
					{
						timer[timer_no] = 0;		// Simulates that the connection attempt timed out
						attempts_left = 0;
					}
				}
				else
				{
					timer[DISCONNECT] = 0xFFFF;
				}

				if(timer[timer_no] == 0)			// If connection attempt has timed out -> abort
				{
					timer[DISCONNECT] = 0xFFFF;
					timer[timer_no] = 0xFFFF;
					comclose(com_opened);
					com_opened = -1;

					if(attempts_left && state == STATE_CONNECTING)
					{
						goodreadled(BLUE_FLASH, -1);
						state = STATE_WAIT_TO_RECONNECT;
						timer[CONNECTING_TIME_OUT] = RECONNECT_WAIT_TIME * 50;
					}
					else
					{
						goodreadled(BLUE_FLASH, 0);
						sound(TVLONG, app.buzzer_volume, SLOW, 0);
						state = STATE_WAIT_FOR_RELEASE;
					}
				}

				break;

			case STATE_CONNECTED_MASTER:			// Connected state
			case STATE_CONNECTED_SLAVE:

				if(blt_is_connected() == FALSE)		// If Bluetooth connection lost
				{
					timer[DISCONNECT] = 0xFFFF;
					goodreadled(BLUE_FLASH, -1);
					sound(TVLONG, app.buzzer_volume, SLOW, 0);
					autopowerdown(OFF, 0);			// Don't turn off while trying to reconnect or being connectable

					if(state == STATE_CONNECTED_MASTER || com_opened == COM12)		// If connected as master, then reconnect as master
					{																// If connected as HID device that was connected as slave (COM12), also reconnect as master (COM13)
						comclose(com_opened);
						com_opened = -1;

						if(app.reconnect_attempts)
						{
							state = STATE_WAIT_TO_RECONNECT;
							timer[CONNECTING_TIME_OUT] = RECONNECT_WAIT_TIME * 50;
							attempts_left = app.reconnect_attempts;
						}
						else
						{
							goodreadled(BLUE_FLASH, 0);
							state = STATE_WAIT_FOR_RELEASE;
						}
					}
					else															// Else -> stay connectable (as slave) and wait for an incomming connection
					{
						state = STATE_CONNECTABLE;
						timer[CONNECTABLE_TIME_OUT] = app.reconnect_attempts * 15 * 50;
					}

					break;
				}

				if(timer[BLUE_FLASH_SLOW] == 0xFFFF)								// Implementation of slow blinking blued LED as connection indicator
				{																	// Note: Connected + sleep mode (powerdown) = no blinking blue LED
					timer[BLUE_FLASH_SLOW] = BLUE_FLASH_SLOW_TIME;
				}
				else if(timer[BLUE_FLASH_SLOW] == 0)
				{
					timer[BLUE_FLASH_SLOW] = BLUE_FLASH_SLOW_TIME;
					goodreadled(BLUE, 10);
				}

				//
				// Wake up iPhones/iPads on trigger press (if enabled)
				//
				if( (app.options & ENABLE_WAKEUP_KEY) && !pressed && triggerpressed() )
				{
					pressed = TRUE;
					putnstring("", 0);	
				}
				else if(!triggerpressed())
				{
					pressed = FALSE;
				}
				
				if(downpressed())			// Clear key can be used to manually disconnect
				{
					if(timer[DISCONNECT] == 0xFFFF && app.trigger_to_disconnect_time != 0)
					{
						timer[DISCONNECT] = app.trigger_to_disconnect_time * 50;
						if( app.options &  ENABLE_KEYBOARD_TOGGLE )
							putcom( 0xA6 );  // toggle iPhone on screen keyboard.
					}
					else if(timer[DISCONNECT] == 0)
					{
						timer[DISCONNECT] = 0xFFFF;
						sound( TVLONG, app.buzzer_volume, SHIGH, SLOW, 0 );
						goodreadled(BLUE_FLASH, 0);
						comclose(com_opened);
						com_opened = -1;
						state = STATE_WAIT_FOR_RELEASE;
					}
				}
				else
				{
					timer[DISCONNECT] = 0xFFFF;
				}

				break;

			case STATE_WAIT_TO_RECONNECT:

				if(triggerpressed())		// Trigger key can be used to do an immediat reconnect attempt
				{
					timer[DISCONNECT] = 0xFFFF;

					if(timer[CONNECT_MASTER] == 0xFFFF && app.trigger_to_connect_time != 0)
					{
						timer[CONNECT_MASTER] = app.trigger_to_connect_time * 50;
					}
					else if(timer[CONNECT_MASTER] == 0)
					{
						timer[CONNECT_MASTER] = 0xFFFF;
						sound(TSTANDARD, app.buzzer_volume, SHIGH, SMEDIUM, 0);
						timer[CONNECTING_TIME_OUT] = 0;				// Forces the waiting time to reconnect to pass immediatly
						attempts_left = 0;
					}
				}
				else if(downpressed())		// Clear key can be used to end the reconnect attempt
				{
					timer[CONNECT_MASTER] = 0xFFFF;

					if(timer[DISCONNECT] == 0xFFFF && app.trigger_to_disconnect_time != 0)
					{
						timer[DISCONNECT] = app.trigger_to_disconnect_time * 50;
					}
					else if(timer[DISCONNECT] == 0)
					{
						timer[DISCONNECT] = 0xFFFF;
						sound( TVLONG, app.buzzer_volume, SHIGH, SLOW, 0 );
						goodreadled(BLUE_FLASH, 0);
						comclose(COM3);
						state = STATE_WAIT_FOR_RELEASE;
						attempts_left = 0;
					}
				}
				else
				{
					timer[CONNECT_MASTER] = 0xFFFF;
					timer[DISCONNECT] = 0xFFFF;
				}

				if(timer[CONNECTING_TIME_OUT] == 0)				// Waiting time to reconnect passed -> try to reconnect
				{
					timer[CONNECTING_TIME_OUT] = 0xFFFF;
					timer[CONNECT_MASTER] = 0xFFFF;
					timer[DISCONNECT] = 0xFFFF;

					goodreadled(BLUE_FLASH, FOREVER);

					com_opened = (app.blt_interface == BLT_HID_DEVICE) ? COM13 : COM3;	// Connect as master as either VCP or HID

					if(comopen(com_opened) == OK)
					{
						state = STATE_CONNECTING;
						timer[CONNECTING_TIME_OUT] = RECONNECTING_TIME * 50;
						autopowerdown(OFF, 0);						// Don't turn off while trying to reconnect

						if(attempts_left)
							--attempts_left;
					}
					else
					{
						goodreadled(BLUE_FLASH, 0);
						comclose(com_opened);
						com_opened = -1;
						sound(TVLONG, app.buzzer_volume, SLOW, 0);
						state = STATE_WAIT_FOR_RELEASE;
						attempts_left = 0;
					}
				}

				break;

			case STATE_WAIT_FOR_RELEASE:		// Waits here until all keys are released before returning to idle state
			default:
				if(!triggerpressed() && !downpressed())
				{
					state = STATE_DISCONNECTED;
					autopowerdown(ON, IDLE_POWER_DOWN_TIME);
				}
				break;
		}


		if(com_opened == -1)		// Makes sure that barcodes can always can be transmitted by USB or Cradle if there's no Bluetooth connection
		{
			com_opened = (app.options & USB_HID_MODE) ? COM10 : COM0;		// Select either USB-HID or USB-VCP
			comopen( com_opened );		
		}

		idle();		// This reduces the power consumption
	}
}
