/*******************************************************************************************************/
//
// File Name      : cmd_handler.c
// Author         : Ronny de Winter
//
/*******************************************************************************************************/
#include <string.h>
#include "lib.h"
#include "RFG3531.h"
#include "cmd_handler.h"

//------------- Defines ------------------------
#define IDLE				0		// Receive states
#define RECEIVING			1
#define HEADER_RECEIVING	2
#define BCC_RECEIVING		3
#define BCC_RECEIVING_L		4
#define EXECUTE_COMMAND		5

#define OSE_CMD_BUF_LEN			200

//-------------- variables ----------------------------
static unsigned int  ose_command_length;
static unsigned char ose_command_buf[OSE_CMD_BUF_LEN+1];
static unsigned int  ose_command_state;

//------------- Public functions -----------------------
void ResetCmdHandler(void);
int CheckSerialCommands(unsigned char ch);

//------------- Local functions ------------------------
static int IsCmdHeader(unsigned int ch);
static int IsCmdTerminator(unsigned int ch);

//------------------------------------------------------------------
//  ResetCmdHandler
//  =====================
//  Resets the variables used for serial command handling
//------------------------------------------------------------------
void ResetCmdHandler(void)
{
	ose_command_state = IDLE;
}

//-----------------------------------------------------------------------------
//  CheckSerialCommands
//  =====================
//  Checks for OSE serial commands and executes them.
//-----------------------------------------------------------------------------
int CheckSerialCommands(unsigned char ch)
{
	ch &= 0x7F;							// Could be communicating with 7bit data

	switch(ose_command_state)
	{
		case IDLE:

			if( IsCmdHeader(ch) )
			{
				ose_command_state = RECEIVING;
				ose_command_length = 0;

				if( IsCmdHeader(ch) == STX || IsCmdHeader(ch) == ESC)		// Only if no header is required jump straight to next state
					break;
			}
			else
				break;

		case RECEIVING:

			if( IsCmdHeader(ch) == STX || IsCmdHeader(ch) == ESC)	// If new header -> start over
			{
				ose_command_length = 0;
			}
			else if ( !IsCmdTerminator(ch) )						// If no terminator -> store char
			{
				ose_command_buf[ose_command_length++] = ch;

				if (ose_command_length >= OSE_CMD_BUF_LEN)
					ose_command_state = IDLE;
			}
			else if(ose_command_length == 0)						// if command was empty -> start over
			{
				ose_command_state = IDLE;
			}
			else													// If terminator received and command is not empty -> execute command
			{
				ose_command_buf[ose_command_length] = '\0';
				ose_command_state = IDLE;

				if( (app.ser_cmd_options & RS_CONFIG_DISABLED)==0 || (ose_command_length == 1))
				{
					if(systemsetting((const char *)ose_command_buf) == OK)
					{
						if(app.ser_cmd_options & RS_ACK_NAK_ON)
							putcom(ACK);
					}
					else
					{
						if(app.ser_cmd_options & RS_ACK_NAK_ON)
							putcom(NAK);
					}

					return TRUE;
				}
			}
			break;
		default:
			ose_command_length = 0;
			ose_command_state = IsCmdHeader(ch) ? RECEIVING : IDLE;
			break;
	}

	return FALSE;
}

//------------------------------------------------------------------
//  IsCmdHeader
//  =====================
//  Checks if a received character is valid command header
//------------------------------------------------------------------
int IsCmdHeader(unsigned int ch)
{
	if( (app.ser_cmd_format & CMD_HEADER_ESC) && (ch == ESC))			// ESC
		return ESC;

	// Check for STX header
	if( (app.ser_cmd_format & CMD_HEADER_STX) && (ch == STX))			// STX
		return STX;

	// Check for No command header
	if( (app.ser_cmd_format & CMD_HEADER_MASK) == 0 )					// No header needed
		return TRUE;

	return FALSE;
}

//------------------------------------------------------------------
//  IsCmdTerminator
//  =====================
//  Checks if a received character is valid command terminator
//------------------------------------------------------------------
int IsCmdTerminator(unsigned int ch)
{
	if( (app.ser_cmd_format & CMD_TERM_CR) && (ch == CR))				// CR
		return CR;

	if( (app.ser_cmd_format & CMD_TERM_LF) && (ch == LF))				// LF
		return LF;

	if( (app.ser_cmd_format & CMD_TERM_ETX) && (ch == ETX))				// ETX
		return ETX;

	if( (app.ser_cmd_format & CMD_TERM_EOT) && (ch == EOT))				// EOT
		return EOT;

	if( (app.ser_cmd_format & CMD_TERM_MASK) == 0 && (ch == CR))		// If no terminator is specified, allow CR
		return CR;

	return FALSE;
}

